
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '32.0.0rc1',
    'branch_nick': 'master',
    'revision_id': '3df51d61a1c2d27564b2fa305e55c74562c9f725',
    'revno': 12582
}
